##################################################################
# November 2017: ADMM for Robust Matrix Regression
# Gamma = 3*lambda
##################################################################
robustmat <- function(Y,X,lambda,gamma,tau,convergence=1e-7,maxiter=1000,rho,init.D=NULL,init.UD=NULL,init.A=NULL,init.Z=NULL,init.W=NULL,init.UZ=NULL,init.UW=NULL){
	
	n <- nrow(Y)
	q <- ncol(Y)
	p <- ncol(X)
		
# Variables Initialization
	D <- UD <-  matrix(0,n,q)
	A <- Z <- W <- matrix(0,p,q)
	oldA <- matrix(1,p,q)
	UZ <- UW <- matrix(0,p,q)	
	criteria <- 1e10
	i <- 1
	
	if(is.null(init.D)==FALSE){
		D <- init.D
		UD <- init.UD		
		A <- init.A
		Z <- init.Z
		W <- init.W
		UZ <- init.UZ
		UW <- init.UW										
	}
	
	tildeX <- rbind(X,diag(1,p,p),diag(1,p,p))
	solvetildeX <- solve(t(tildeX)%*%tildeX)
	
# While loop for the iterations
	while(criteria > convergence && i <= maxiter){
	
	A <- updateA(solvetildeX,tildeX,D,Z,W,UD,UZ,UW)
		
	Z <- updateZ(A,UZ,lambda,gamma,rho)
	
	W <- updateW(A,UW,lambda,rho)

	D <- updateD(Y,X,A,UD,rho,tau)
		
	UD <- UD+D-X%*%A
	
	UZ <- UZ+Z-A
	
	UW <- UW+W-A		
	
	criteria <- mean((A-oldA)^2)
#	print(criteria)
#	if(sum(abs(oldA))<=1e-8){ return(list(A=A,iteration=i))}
	oldA <- A
	i <- i+1
		
	}	
	
	if(i>maxiter){
  	 warning("The algorithm has not converged by the specified maximum number of iteration")
  }

return(list(A=A,Z=Z,W=W,D=D,UD=UD,UZ=UZ,UW=UW,iteration=i))	
	
}


######################################
# Cross-validation
######################################
cv.robust <- function(X,Y,lambdas,gamma,nfold,loss="huber",seed,tau=3){
	
	set.seed(seed)
    p <- ncol(X)
    q <- ncol(Y) 
# Index for training vs test set
	fold <- rep(NA,nrow(Y))
	fold <- sample(rep(1:nfold,nrow(Y)/nfold))
	cv.error <- NULL
	
for(j in 1:nfold){
		print(paste("cross validation for dataset ",j,sep=""))
		error <- NULL
		tmp <- 1
		ytrain <- Y[which(fold!=j),]
		ytest <- Y[which(fold==j),]		
		Xtrain <- X[which(fold!=j),]		
		Xtest <- X[which(fold==j),]			
		ntrain <- nrow(Xtrain)
		if(loss=="huber"){
			tau <- tau
		}
		else if(loss=="sqerror"){
			tau <- 100000000
		}		
	init.D <- init.UD <-  matrix(0,nrow(Xtrain),q)
	init.A <- init.Z <- init.W <- matrix(0,p,q)
	init.UZ <- init.UW <- matrix(0,p,q)	


		for(lambda in lambdas){
		
		gammaupdate <- gamma*lambda
		model <- robustmat(ytrain,Xtrain,lambda,gammaupdate,tau,convergence=1e-6,maxiter=1000,rho=10/nrow(Xtrain),init.D=init.D,init.UD=init.UD,init.A=init.A,init.Z=init.Z,init.W=init.W,init.UZ=init.UZ,init.UW=init.UW)
		init.D <- model$D
		init.UD <- model$UD
		init.A <- model$A
		init.Z <- model$Z
		init.W <- model$W
		init.UZ <- model$UZ
		init.UW <- model$UW
				
		estA <- model$A
		predictedY <- Xtest%*%estA
		error <- c(error,sum((predictedY-ytest)^2))	
}
		cv.error <- rbind(cv.error,error)
	}
	
	return(list(cv.error=cv.error,lambdas=lambdas,gamma=gamma))
}


######################################################
# Update A
######################################################
updateA <- function(solvetildeX,tildeX,D,Z,W,UD,UZ,UW){
	Theta <- rbind(D,Z,W)
	U <- rbind(UD,UZ,UW)
	return(solvetildeX%*%t(tildeX)%*%(Theta+U))
	
}

######################################################
# Update Z
######################################################
updateZ <- function(A,UZ,lambda,gamma,rho){
  temp <- A-UZ
  B <- lambda*gamma/rho
  S <- Soft(temp,B)
  return(S)	
}


######################################################
# Update D
######################################################
updateD <- function(Y,X,A,UD,rho,tau){
  n <- nrow(X)
  q <- ncol(A)
  C <- X%*%A-UD
  D <- matrix(0,n,q)
  index <- which(abs(n*rho*(Y-C)/(1+n*rho)) <= tau)
#  index2 <- which(abs(C)<= tau*(1+n*rho)/(n*rho))
  D[index] <- (Y + n*rho*C)[index]/(1+n*rho)
  D[-index] <- Y[-index] - Soft(Y-C,tau/(n*rho))[-index]
  return(D)	
}


######################################################
# Update W
######################################################
updateW <- function(A,UW,lambda,rho){
  C <- A-UW
  a <- svd(C)
  D <- diag(a$d)
  U <- a$u
  V <- a$v  
  L <- U%*%(pmax(D-lambda/rho,0))%*%t(V)			
  return(L)
}


######################################################
# Soft-thresholding Operator
######################################################
Soft <- function(a,b){
  if(b<0) stop("Can soft-threshold by a nonnegative quantity only.")
  return(sign(a)*pmax(0,abs(a)-b))
}






